/*------------------------------------------------------------------------*\
 * Copyright (C) 2015 - 2016 Armin Biere Johannes Kepler University Linz  *
\*------------------------------------------------------------------------*/

static const char * USAGE =
"usage: drabt [ <option> ... ] <pre.cnf> [ <drat.proof> [ <post.cnf> ] ]\n"
"\n"
"-h    print this usage information\n"
"      (also printed if <pre.cnf> is missing)\n"
"\n"  
"-f    complete forward checking (default)\n"
"-b    partial backward checking (not implemented yet)\n"
"-p    enable plain mode and ignore all 'd' lines\n"
"-S    run in forward SAT mode (do not necessarily check for refutation)\n"
"\n"
"-v                   increase verbose level\n"
"                     (default 0, compiled in maximum %d)\n"
"\n"                 
"-M | --more-details  print more details in error messages (line numbers)\n"
"                     (needs more memory, default in verbose mode)\n"
"\n"                 
"-L | --less-details  print less details in error messages\n"
"                     (needs less memory, default in non-verbose mode)\n"
"\n"
"-I | --ignore-warnings   ignore and do not even print warnings\n"
"\n"
"-P | --pedantic      turns warnings into errors\n"
"                     (for instance abort if 'p' header is missing)\n"
"\n"                 
"-s | --sloppy        ignore certain (ignorable) errors\n"
"                     (like 'd' clauses which do not exist anymore)\n"
"\n"
"--force-non-binary   assume proof is in non-binary format\n"
"\n"
"<pre.cnf>      antecedent or input CNF in DIMACS format\n"
"<drat.proof>   DRAT proof showing <post.cnf> is implied by <pre.cnf>\n"
"<post.cnf>     consequent or implied CNF in DIMACS format\n"
"\n"
"If <post.cnf> is omitted then the checker assumes that <post.cnf> just\n"
"contains the empty clause, which means that <drat.proof> is supposed\n"
"to be a DRAT refutation for <pre.cnf>.  If the file name <drat.proof>\n"
"is missing then the checker reads the DRAT proof file from <stdin>.\n"
"\n"
"The tool has support for reading compressed CNFS and proofs in 'gzip',\n"
"'zip', 'bzip2', 'xz' and '7z' format, but relies on external tools to\n"
"do so, actually using pipes through the 'popen' command.  Decompression is\n"
"triggered if a given file name has a corresponding suffix, i.e., '.gz',\n"
"'.zip', 'bz2', '.xz' or '.7z' and thus of course does not work when\n"
"reading from <stdin>.\n"
"\n"
"The proof is checked to be a valid DRAT proof of all the clauses in\n"
"<post.cnf> starting with the clauses in <pre.cnf>.  Since a DRAT proof\n"
"also contains deletion information, the clauses in <post.cnf> must not\n"
"only occur in the proof but should also still be present at the end of\n"
"the proof.  Further, not all clauses still alive at the end of the proof\n"
"have to be in <post.cnf>.\n"
"\n"
"If the check succeeds 's VERIFIED' is printed on a single line.  In this\n"
"case the checker will produce a zero exit code.  If the check fails the\n"
"reason for the failure is explained as error message and the program\n"
"exits with a non zero exit code.\n"
;

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/

static void die (const char * fmt, ...) {
  va_list ap;
  fflush (stdout);
  fputs ("*** drabt error: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static int verbosity, less_details, more_details, details;

#if MAXVERBOSITY > 0

static void msg (int level, const char * fmt, ...) {
  va_list ap;
  if (verbosity < level) return;
  fputs ("c ", stdout);
  va_start (ap, fmt);
  vfprintf (stdout, fmt, ap);
  va_end (ap);
  fputc ('\n', stdout);
  fflush (stdout);
}

#define msg1(ARGS...) msg (1, ##ARGS)

#else
#define msg1(ARGS...) do { } while (0)
#endif

#if MAXVERBOSITY > 1
#define msg2(ARGS...) msg (2, ##ARGS)
#define cmsg2(ARGS...) cmsg (2, ##ARGS)
#else
#define msg2(ARGS...) do { } while (0)
#define cmsg2(ARGS...) do { } while (0)
#endif

#if MAXVERBOSITY > 2
#define msg3(ARGS...) msg (3, ##ARGS)
#else
#define msg3(ARGS...) do { } while (0)
#endif

/*------------------------------------------------------------------------*/

#if MAXVERBOSITY > 0

static double percent (double a, double b) { return b ? 100.0 * a / b : 0; }

static double average (double a, double b) { return b ? a / b : 0; }

#endif

/*------------------------------------------------------------------------*/

#if MAXVERBOSITY > 0
typedef struct Timer { int level; double process, real; } Timer;

static double get_process_time (void) {
  struct rusage u;
  double res;
  if (getrusage (RUSAGE_SELF, &u)) return 0;
  res = u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
  res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
  return res;
}

static double start_real_time;

static double current_real_time (void) {
  double res = 0;
  struct timeval tv;
  if (!gettimeofday (&tv, 0)) res = 1e-6 * tv.tv_usec, res += tv.tv_sec;
  return res;
}

static double get_real_time (void) {
  assert (start_real_time);
  return current_real_time () - start_real_time;
}

static void start_timer (int level, Timer * timer) {
  if (verbosity < (timer->level = level)) return;
  timer->process = get_process_time ();
  timer->real = get_real_time ();
}

static void stop_timer (Timer * timer, const char * fmt, ...) {
  va_list ap;
  if (verbosity < timer->level) return;
  timer->process = get_process_time () - timer->process;
  timer->real = get_real_time () - timer->real;
  fputs ("c ", stdout);
  va_start (ap, fmt);
  vfprintf (stdout, fmt, ap);
  va_end (ap);
  printf (" %.2f sec (%.0f%% of real time %.2f sec)",
    timer->process, percent (timer->process, timer->real), timer->real);
  fputc ('\n', stdout);
  fflush (stdout);
}

#endif

/*------------------------------------------------------------------------*/

static size_t allocated, maxbytes;

static void inc_allocated (size_t bytes) {
  allocated += bytes;
  if (allocated > maxbytes) maxbytes = allocated;
}

static void dec_allocated (size_t bytes) {
  assert (allocated >= bytes);
  allocated -= bytes;
}

static void * allocate (size_t bytes) {
  void * res = malloc (bytes);
  if (!res) die ("out of memory allocating %ld bytes", (long) bytes);
  inc_allocated (bytes);
  memset (res, 0, bytes);
  return res;
}

static void deallocate (void * ptr, size_t bytes) {
  dec_allocated (bytes);
  free (ptr);
}

static void *
reallocate (void * ptr, size_t old_bytes, size_t new_bytes) {
  char * res;
  dec_allocated (old_bytes);
  res = realloc (ptr, new_bytes);
  if (!res)
    die ("out of memory reallocating %ld to %ld bytes",
      old_bytes, new_bytes);
  inc_allocated (new_bytes);
  if (new_bytes > old_bytes)
    memset (res + old_bytes, 0, new_bytes - old_bytes);
  return res;
}

/*------------------------------------------------------------------------*/

#define NEW(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = allocate (BYTES); \
} while (0)

#define DEL(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  deallocate ((P), BYTES); \
  (P) = 0; \
} while (0)

#define RSZ(P,O,N) \
do { \
  size_t OLD_BYTES = (O) * sizeof *(P); \
  size_t NEW_BYTES = (N) * sizeof *(P); \
  (P) = reallocate ((P), OLD_BYTES, NEW_BYTES); \
} while (0)

/*------------------------------------------------------------------------*/

#define STACK(TYPE) struct { TYPE * start; TYPE * top; TYPE * end; }

#define EMPTY(S) ((S).top == (S).start)
#define FULL(S) ((S).top == (S).end)

#define COUNT(S) ((S).top - (S).start)
#define SIZE(S) ((S).end - (S).start)

#define ENLARGE(S) \
do { \
  size_t OLD_SIZE = SIZE (S); \
  size_t NEW_SIZE = OLD_SIZE ? 2*OLD_SIZE : 1; \
  size_t OLD_COUNT = COUNT (S); \
  RSZ ((S).start, OLD_SIZE, NEW_SIZE); \
  (S).top = (S).start + OLD_COUNT; \
  (S).end = (S).start + NEW_SIZE; \
} while (0)

#define PUSH(S,E) \
do { \
  if (FULL(S)) ENLARGE (S); \
  *(S).top++ = (E); \
} while (0)

#define PEEK(S,I) \
  (assert ((I) < COUNT (S)), (S).start [(I)])

#define RESET(S) \
do { \
  (S).start = (S).top = (S).end = 0; \
} while (0)

#define CLEAR(S) \
do { \
  (S).top = (S).start; \
} while (0)

#define RELEASE(S) \
do { \
  size_t N = SIZE(S); \
  DEL ((S).start, N); \
  RESET(S); \
} while (0)

/*------------------------------------------------------------------------*/

static int close_input_file;
static unsigned long lineno, charno, start_of_token, readchars;
static unsigned long position_at_start_of_clause;
static const char * input_file_name;
static FILE * input_file;

static int has_suffix (const char * str, const char * suffix) {
  int l = strlen (str), k = strlen (suffix);
  return (l >= k) && !strcmp (str + l - k, suffix);
}

static void pipe_cmd (const char * name, const char * fmt) {
  int len = strlen (name) + strlen (fmt) + 1;
  char * cmd;
  NEW (cmd, len);
  sprintf (cmd, fmt, name);
  msg1 ("piping '%s' through '%s'", name, cmd);
  input_file = popen (cmd, "r");
  if (!input_file) die ("can not open pipe '%s' for reading", cmd);
  DEL (cmd, len);
  close_input_file = 2;
  input_file_name = name;
}

static const char * stdin_name = "<stdin>";

static void open_input_file_for_reading (const char * name) {
  if (!name) {
    input_file = stdin;
    input_file_name = stdin_name;
    close_input_file = 0;
  } else if (has_suffix (name, ".gz")) pipe_cmd (name, "gunzip -c %s");
  else if (has_suffix (name, ".xz")) pipe_cmd (name, "xz -c -d %s");
  else if (has_suffix (name, ".bz2")) pipe_cmd (name, "bzcat %s");
  else if (has_suffix (name, ".zip")) pipe_cmd (name, "unzip -p %s");
  else if (has_suffix (name, ".7z"))
    pipe_cmd (name, "7z x -so %s 2>/dev/null");
  else if (!(input_file = fopen (name, "r")))
    die ("can not open file '%s' for reading", name);
  else input_file_name = name, close_input_file = 1;
}

static STACK(char) line;
static unsigned long end_of_line;
static int last_end_of_line_char;

static STACK(int) buffer;
static unsigned long end_of_buffer;
static int buffering;

static int buffered_read_next_char () {
  int res;
  if (buffering) {
    res = getc_unlocked (input_file);
    PUSH (buffer, res);
  } else if (EMPTY (buffer))
    res = getc_unlocked (input_file);
  else if (end_of_buffer < COUNT (buffer)) 
    res = buffer.start[end_of_buffer++];
  else {
    res = getc_unlocked (input_file);
    end_of_buffer = 0;
    RELEASE (buffer);
  }
  return res;
}

static int read_next_binary_char () {
  int ch = buffered_read_next_char ();
  if (ch != EOF) readchars++;
  return ch;
}

static int read_next_line () {
  int ch = buffered_read_next_char ();
  if (ch == EOF) return 0;
  CLEAR (line);
  while (ch != '\n') {
    if (ch == EOF) break;
    PUSH (line, ch);
    ch = buffered_read_next_char ();
  }
  PUSH (line, 0);
  last_end_of_line_char = ch;
  end_of_line = COUNT (line);
  charno = 0;
  lineno++;
  return 1;
}

static int read_next_char () {
  int ch;
  assert (charno <= end_of_line);
  if (charno == end_of_line && !read_next_line ()) return EOF;
  assert (charno < COUNT (line));
  ch = line.start[charno++];
  if (charno == end_of_line) assert (!ch), ch = last_end_of_line_char;
  readchars++;
  return ch;
}

static void save_char (int ch) {
  assert (0 < charno);
  assert (readchars > 0), readchars--;
  charno--;
}

static void print_line_and_cursor () {
  unsigned long actual_start_of_line, actual_end_of_line;
  unsigned long actual_start_of_token;
  long i, chars_to_print;
  char marker;
  if (EMPTY (line)) return;
  if (!input_file) return;
  assert (end_of_line == COUNT (line));
  assert (start_of_token <= charno);
  chars_to_print = charno - start_of_token;
  if (chars_to_print < 74) actual_start_of_line = 0;
  else actual_start_of_line = start_of_token - 70;
  actual_end_of_line = actual_start_of_line + 66;
  if (actual_end_of_line > end_of_line) actual_end_of_line = end_of_line;
  for (i = actual_start_of_line; i < actual_end_of_line; i++)
    if (isprint ((int) PEEK (line, i)))
      break;
  if (i == actual_end_of_line) {
    assert (actual_start_of_line < actual_end_of_line);
    fputs (" ... <only-non-printable-characters-in-this-line> ...\n", stderr);
    return;
  }
  if (actual_start_of_line) fputs ("... ", stderr);
  else fputc (' ', stderr);
  for (i = actual_start_of_line;
       i < actual_end_of_line && i < end_of_line - 1;
       i++) {
    int ch = PEEK (line, i);
    fputc ((isprint (ch) ? ch : '@'), stderr);
  }
  if (actual_end_of_line < end_of_line) fputs (" ... ", stderr);
  fputc ('\n', stderr);
  if (actual_start_of_line) fputs ("    ", stderr);
  for (i = actual_start_of_line; i < start_of_token; i++)
    fputc (' ', stderr);
  actual_start_of_token =
    (start_of_token < actual_start_of_line ?
      actual_start_of_line : start_of_token);
  marker = (charno - actual_start_of_line > 1) ? '~' : ' ';
  for (i = actual_start_of_token; i < charno; i++)
    fputc (marker, stderr);
  fputc ('^', stderr);
  fputc ('\n', stderr);
}

static struct Context * current_context;
static const char * context_type (struct Context *);

static void perr (const char * fmt, ...) {
  va_list ap;
  fflush (stdout);
  fprintf (stderr, "%s:", input_file_name);
  if (input_file) fprintf (stderr, "%lu:%lu: ", lineno, charno);
  else fputc (' ', stderr);
  fprintf (stderr, "error in %s: ", context_type (current_context));
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  print_line_and_cursor ();
  fflush (stderr);
  exit (1);
}

static void berr (const char * fmt, ...) {
  va_list ap;
  fflush (stdout);
  fprintf (stderr, "%s:", input_file_name);
  if (input_file) fprintf (stderr, " byte %lu: ", readchars);
  else fputc (' ', stderr);
  fprintf (stderr, "error in %s: ", context_type (current_context));
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static int pedantic, ignore_warnings, sloppy;

static void pwrn (const char * fmt, ...) {
  va_list ap;
  if (ignore_warnings) return;
  fflush (stdout);
  fprintf (stderr, "%s:", input_file_name);
  if (input_file) fprintf (stderr, "%lu:%lu: ", lineno, charno);
  else fputc (' ', stderr);
  fprintf (stderr, "warning");
  if (pedantic) fprintf (stderr, " treated as error");
  fprintf (stderr, ": ");
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  print_line_and_cursor ();
  fflush (stderr);
  if (pedantic) exit (1);
}

static void close_input_file_if_necessary () {
  if (close_input_file == 1 && fclose (input_file))
     pwrn ("failed to close file '%s'", input_file_name);
  if (close_input_file == 2 && pclose (input_file))
     pwrn ("failed to close pipe to read '%s'", input_file_name);
  msg1 ("closed '%s' after reading %lu bytes %.1f MB",
    input_file_name, readchars, readchars/(double)(1ul<<20));
  close_input_file = 0;
  input_file = 0;
}

static int is_white_space (int ch) {
  return ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r';
}

static char printable_description_of_char_buffer[80];

static const char * printable_description_of_char (int ch) {
       if (ch == EOF)  return "end-of-file";
  else if (ch == '\n') return "new-line '\\n'";
  else if (ch == '\t') return "tab '\\t'";
  else if (ch == ' ')  return "space ' '";
  else if (isprint (ch)) {
    sprintf (printable_description_of_char_buffer, "'%c'", ch);
    return printable_description_of_char_buffer;
  } else {
    sprintf (printable_description_of_char_buffer,
      "non-printable character code 0x%02x", ch);
    return printable_description_of_char_buffer;
  }
}

static char but_got_buffer[80];

static const char * but_got (int ch) {
  sprintf (but_got_buffer,
    " but got %s", printable_description_of_char (ch));
  return but_got_buffer;
}

static int parse_non_negative_int (const char * type) {
  int ch = read_next_char (), res;
  if (ch == '0') {
    ch = read_next_char ();
    res = 0;
  } else if (!isdigit (ch))
    perr ("expected digit at start of %s", type);
  else {
    res = ch - '0';
    while (isdigit (ch = read_next_char ())) {
      int digit;
      if (INT_MAX/10 < res)
	perr ("%s exceeds INT_MAX substantially", type);
      res *= 10;
      digit = ch - '0';
      if (INT_MAX - digit < res)
	perr ("%s exceeds INT_MAX", type);
      res += digit;
    }
  }
  save_char (ch);
  return res;
}

/*------------------------------------------------------------------------*/

typedef struct Cls {
  struct Cls * chain;
  unsigned long refs, hash;
#ifndef NCHECK
  unsigned long watched;
#endif
  unsigned used:1, satisfied:1, empty:1, size;
  int lits[2];
} Cls;

typedef STACK(Cls*) ClsPtrStk;

typedef struct Context {
  int type;		// -1 = pre, 0 = proof, 1 = post
  const char * name;
  int read_from_stdin, binary;
  struct { unsigned long idx; int found; } empty;
  struct { int maxvar, clauses, found; } header;
  STACK(unsigned long) position;
  STACK(char) deleted;
  ClsPtrStk clauses;
} Context;

typedef struct Position {
  Context * context;
  unsigned long idx;
} Position;

typedef struct Details {
  unsigned long id;
  Position first, add, del;
} Details;

typedef struct Buffer {
  char * start;
  size_t size;
} Buffer;

/*------------------------------------------------------------------------*/

static size_t bytes_clause (int size) {
  int additional_lits = (size < 3) ? 0 : size - 2;
  size_t res = sizeof (Cls) + additional_lits * sizeof (int);
  if (details) res += sizeof (Details);
  return res;
}

static Cls ** clause_hash_table;
static unsigned long size_clause_hash_table;
static unsigned long num_entries_clause_hash_table;

#if MAXVERBOSITY > 0
static unsigned long searches_clause_hash_table;
static unsigned long collisions_clause_hash_table;
static unsigned long checks, decisions, propagations;
#define INC(STATS) do { (STATS) += 1; } while (0)
#else
#define INC(STATS) do { } while (0)
#endif

static void enlarge_clause_hash_table () {
  unsigned long old_size = size_clause_hash_table, new_size;
  Cls ** new_table, * c, * n, ** p;
  unsigned long i, hash, mask;
  new_size = old_size ? 2*old_size : 1;
  mask = new_size - 1;
  NEW (new_table, new_size);
  for (i = 0; i < old_size; i++) {
    for (c = clause_hash_table[i]; c; c = n) {
      n = c->chain;
      hash = c->hash & mask;
      p = new_table + hash;
      c->chain = *p;
      *p = c;
    }
  }
  DEL (clause_hash_table, old_size);
  size_clause_hash_table = new_size;
  clause_hash_table = new_table;
}

static int cmp_lits_for_qsort (const void * p, const void * q) {
  int a = *(int *) p, b = *(int *) q;
  int res = abs (a) - abs (b);
  if (res) return res;
  return a - b;
}

static STACK(int) clause;
static unsigned long clause_hash, clause_size;
static unsigned long clauses_with_duplicated_literals;

static void remove_duplicates () {
  int * q = clause.start, lit, prev = 0, duplicate = 0;
  const int * p;
  for (p = q; p < clause.top; p++) {
    lit = *p;
    if (prev != lit) *q++ = lit, prev = lit;
    else if (!duplicate) duplicate = lit;
  }
  if (!duplicate) return;
  assert (q < p);
  clause.top = q;
  clause_size = COUNT (clause);
  if (ignore_warnings) return;
  if (!EMPTY (line)) {
    const char * found;
    char dupstr[12];
    sprintf (dupstr, "%d", duplicate);
    found = strstr (line.start, dupstr);
    if (found && (found = strstr (found + 1, dupstr)))
      start_of_token = found - line.start + 1,
      charno = start_of_token + strlen (dupstr) - 1;
  }
  if (++clauses_with_duplicated_literals == 1)
    pwrn ("found clause with duplicated literal %d", duplicate);
  else if (clauses_with_duplicated_literals == 2)
    pwrn ("found another clause with duplicated literal %d (no more warnings)", duplicate);
  charno = end_of_line;
}

static void sort_clause () {
  clause_size = COUNT (clause);
  qsort (clause.start, clause_size, sizeof (int), cmp_lits_for_qsort);
}

static unsigned long primes[] =
{
  123123137, 100000007, 777777799, 543216337, 787945577, 444323447,
  999999929, 234123469, 456789889, 665541251, 564564659, 987019109,
};

static const unsigned num_primes = sizeof primes / sizeof primes[0];

static void hash_clause () {
  const int * p;
  unsigned i = 0;
  clause_hash = 0;
  for (p = clause.start; p < clause.top; p++) {
    unsigned long tmp = clause_hash & 0xffff000000000000ul;
    clause_hash <<= 16;
    clause_hash += primes[i++] * (unsigned long) *p;
    if (tmp) clause_hash ^= tmp >> 48;
    if (i == num_primes) i = 0;
  }
}

static int clause_matches_existing_clause (const Cls * c) {
  const int * p, * q;
  assert (clause_size == COUNT (clause));
  if (clause_size != c->size) return 0;
  if (clause_hash != c->hash) return 0;
  for (q = c->lits, p = clause.start; p < clause.top; p++, q++)
    if (*p != *q) return 0;
  return 1;
}

static Cls ** find_clause_position_in_hash_table () {
  unsigned long h;
  Cls ** res, * c;
  if (size_clause_hash_table == num_entries_clause_hash_table)
    enlarge_clause_hash_table ();
  sort_clause ();
  remove_duplicates ();
  hash_clause ();
  assert (size_clause_hash_table > 0);
  h = clause_hash & (size_clause_hash_table - 1);
  INC (searches_clause_hash_table);
  for (res = clause_hash_table + h;
       (c = *res) && !clause_matches_existing_clause (c);
       res = &c->chain)
    INC (collisions_clause_hash_table);
  return res;
}

/*------------------------------------------------------------------------*/

static Position position (Context * context, unsigned long idx) {
  Position res;
  res.context = context;
  res.idx = idx;
  return res;
}

static int cmp_position (Position a, Position b) {
  int res = (a.context != 0) - (b.context != 0);
  if (res) return res;
  if (a.context) {
    assert (b.context);
    res = a.context->type - b.context->type;
    if (res) return res;
  }
  if (a.idx < b.idx) return -1;
  if (a.idx > b.idx) return 1;
  return 0;
}

static Details * clause_details (Cls * cls) {
  char * p = (char *) cls;
  assert (details);
  p -= sizeof (Details);
  return (Details*) p;
}

static const char * step_type (Context * context, unsigned long idx) {
  if (context->type) return "clause";
  else if (PEEK (context->deleted, idx)) return "deleted clause";
  else return "added clause";
}

static const char * context_type (Context * context) {
  if (context->type < 0) return "pre-CNF";
  else if (context->type > 0) return "post-CNF";
  else return "proof";
}

static void adjust_buffer (Buffer * buffer, size_t required_size) {
  size_t new_size;
  if (buffer->size >= required_size) return;
  for (new_size = 1; new_size < required_size; new_size *= 2)
    ;
  RSZ (buffer->start, buffer->size, new_size);
  buffer->size = new_size;
}

static void release_buffer (Buffer * buffer) {
  DEL (buffer->start, buffer->size);
}

static Buffer file_info_buffer;
static char clause_info_buffer[80];

static const char * file_info (Context * context, unsigned long idx) {
  if (details) {
    sprintf (file_info_buffer.start,
      "%s: %s %lu:",
      context->name,
      (context->binary ? "byte" : "line"),
      PEEK (context->position, idx));
  } else
    sprintf (file_info_buffer.start,
      "%s:",
      context->name);
  return file_info_buffer.start;
}

static const char * clause_info (Context * context, unsigned long idx) {
  if (details) {
    Cls * cls = PEEK (context->clauses, idx);
    Details * d = clause_details (cls);
    sprintf (clause_info_buffer,
      "%s %lu (id %lu) in %s",
      step_type (context, idx), idx+1, d->id, context_type (context));
  } else
    sprintf (clause_info_buffer,
      "%s %lu in %s",
      step_type (context, idx), idx+1, context_type (context));
  return clause_info_buffer;
}

static void check_empty (Context * context, unsigned long idx) {
  Cls * c = PEEK (context->clauses, idx);
  if (c->size) return;
  if (context->empty.found) return;
  msg1 ("%s %s is an empty clause",
    file_info (context, idx), clause_info (context, idx));
  context->empty.found = 1;
  context->empty.idx = idx;
}

static int maxvar;
static unsigned long clause_ids;

static Cls * new_clause (Context * context) {
  const unsigned size = COUNT (clause);
  const size_t bytes = bytes_clause (size);
  char * p = allocate (bytes), * q = p;
  const unsigned long idx = COUNT (context->clauses);
  Details * d;
  unsigned i;
  Cls * res;
  assert (COUNT (clause) <= (long) UINT_MAX);
  if (details) q += sizeof (*d);
  res = (Cls*) q;
  for (i = 0; i < size; i++) {
    int lit = clause.start[i], idx = abs (lit);
    res->lits[i] = lit;
    if (idx > maxvar) maxvar = idx;
  }
  res->size = size;
  res->hash = clause_hash;
  res->refs = 1;
  clause_ids++;
  PUSH (context->clauses, res);
  if (!details) return res;
  d = clause_details (res);
  assert (p == (char*) d);
  d->id = clause_ids;
  d->first = d->add = position (context, idx);
  assert (!d->del.context);
  PUSH (context->position, position_at_start_of_clause);
  return res;
}

static void old_clause (Context * context, Cls * c) {
  PUSH (context->clauses, c);
  if (details) PUSH (context->position, position_at_start_of_clause);
}

static void delete_clause (Cls * c) {
  size_t bytes = bytes_clause (c->size);
  char * p = (char *) c;
  if (details) p -= sizeof (Details);
  deallocate (p, bytes);
}

/*------------------------------------------------------------------------*/

#if MAXVERBOSITY > 1

static void cmsg (int level, 
                  Context * context, unsigned long idx,
		  const char * fmt, ...) {
  va_list ap;
  unsigned i;
  if (verbosity < level) return;
  fputs ("c ", stdout);
  fputs (file_info (context, idx), stdout);
  fputc (' ', stdout);
  fputs (clause_info (context, idx), stdout);
  fputc (' ', stdout);
  va_start (ap, fmt);
  vprintf (fmt, ap);
  va_end (ap);
  if (details) {
    Cls * c = PEEK (context->clauses, idx);
    for (i = 0; i < c->size; i++)
      printf (" %d", c->lits[i]);
    fputs (" 0", stdout);
  }
  fputc ('\n', stdout);
  fflush (stdout);
}

#endif

static void cdie (int ignorable,
                  Context * context, unsigned long idx,
                  const char * fmt, ...) {
  va_list ap;
  unsigned i;
  fflush (stdout);
  fputs (file_info (context, idx), stderr);
  fputc (' ', stderr);
  if (ignorable && sloppy) fputs ("ignorable", stderr);
  fprintf (stderr, "error in %s: ", context_type (context));
  fputs (clause_info (context, idx), stderr);
  fputc (' ', stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  if (details) {
    Cls * c = PEEK (context->clauses, idx);
    Details * d = clause_details (c);
    Position pos = position (context, idx);
    for (i = 0; i < c->size; i++)
      fprintf (stderr, " %d", c->lits[i]);
    fputs (" 0\n", stderr);
    if (!cmp_position (d->first, pos))
      fprintf (stderr, "(this is the first occurrence of this clause)\n");
    else {
      int first_is_add = !cmp_position (d->first, d->add);
      fprintf (stderr,
	"(first %stime added in %s '%s' at line %lu",
	(first_is_add ? "and last " : ""),
	context_type (d->first.context), 
	d->first.context->name,
	PEEK (d->first.context->position, d->first.idx));
      if (!first_is_add)
	fprintf (stderr,
	  ",\n resurrected last time in %s '%s' at line %lu",
	  context_type (d->add.context), 
	  d->add.context->name,
	  PEEK (d->add.context->position, d->add.idx));
      if (d->del.context)
	fprintf (stderr,
	  ",\n last time deleted in %s '%s' at line %lu",
	  context_type (d->del.context), 
	  d->del.context->name,
	  PEEK (d->del.context->position, d->del.idx));
      fputs (")\n", stderr);
    }
  } else if (less_details)
    fputs (
      "(use '-M' or '-v' instead of '-L' for details like line numbers)\n",
      stderr);
  else
    fputs (
      "(use '-M' or '-v' for details like line numbers)\n",
      stderr);
  if (!ignorable || !sloppy) exit (1);
}

/*------------------------------------------------------------------------*/

static Context pre, proof, post;

static void found_pre_cnf_header (int maxvar, int clauses) {
  assert (EMPTY (pre.clauses));
  assert (!pre.header.found);
  assert (maxvar >= 0);
  assert (clauses >= 0);
  pre.header.found = 1;
  pre.header.maxvar = maxvar;
  pre.header.clauses = clauses;
}

static void found_pre_cnf_clause (int deleted) {
  unsigned long idx;
  Cls * c, ** p;
  if (deleted) perr ("found 'd' clause in pre-CNF");
  idx = COUNT (pre.clauses);
  if (!(c = *(p = find_clause_position_in_hash_table ()))) {
    *p = c = new_clause (&pre);
    assert (c->refs == 1);
    num_entries_clause_hash_table++;
    assert (num_entries_clause_hash_table);
  } else {
    old_clause (&pre, c);
    c->refs++, assert (c->refs);
    if (c->refs == 1 && details)
      clause_details (c)->add = position (&pre, idx);
  }
  cmsg2 (&pre, idx, "parsed");
  check_empty (&pre, idx);
}

#define CONTEXT pre
#define PARSE parse_pre_cnf
#define HEADER found_pre_cnf_header
#define CLAUSE found_pre_cnf_clause
#define BINARY 0
#include "parse.h"

/*------------------------------------------------------------------------*/

static int plain_mode;

static void found_proof_header (int nvars, int nclauses) {
  if (charno > 1) charno--;
  pwrn ("CNF header 'p cnf %d %d' in DRAT proof", nvars, nclauses);
}

static void found_proof_clause (int deleted) {
  unsigned long idx;
  Cls * c, ** p;
  idx = COUNT (proof.clauses);
  if (!(c = *(p = find_clause_position_in_hash_table ()))) {
    *p = c = new_clause (&proof);
    PUSH (proof.deleted, deleted);
    if (deleted) cdie (0, &proof, idx, "does not exist (yet)");
    assert (c->refs == 1);
    num_entries_clause_hash_table++;
    assert (num_entries_clause_hash_table);
  } else {
    old_clause (&proof, c);
    PUSH (proof.deleted, deleted);
    if (!deleted) {
      c->refs++, assert (c->refs);
      if (c->refs == 1 && details)
	clause_details (c)->add = position (&proof, idx);
    } else if (!c->refs) {
      assert (!plain_mode);
      cdie (1, &proof, idx, "does not exist anymore");
    } else if (!plain_mode && !--c->refs && details)
      clause_details (c)->del = position (&proof, idx);
  }
  cmsg2 (&proof, idx, "parsed");
  check_empty (&proof, idx);
}

static void parse_binary_proof (int * parsed_clauses_ptr) {
  int first, ch, lit, parsed_clauses = 0;
  unsigned ulit;
  for (;;) {
    first = read_next_binary_char ();
    if (first == EOF) break;
    position_at_start_of_clause = readchars;
    if (first != 'a' && first != 'd') berr ("expected 'a' or 'd'");
    for (;;) {
      ch = read_next_binary_char ();
      if (ch == EOF) berr ("1st byte end-of-file");
      ulit = ch & 0x7f;
      if (ch & 0x80) {
	ch = read_next_binary_char ();
	if (ch == EOF) berr ("2nd byte end-of-file");
	if (!ch) berr ("2nd byte unexpected byte code 0x00");
	ulit |= (ch & 0x7f) << 7;
	if (ch & 0x80) {
	  ch = read_next_binary_char ();
	  if (ch == EOF) berr ("3rd byte end-of-file");
	  if (!ch) berr ("3rd byte unexpected byte code 0x00");
	  ulit |= (ch & 0x7f) << 14;
	  if (ch & 0x80) {
	    ch = read_next_binary_char ();
	    if (ch == EOF) berr ("4th byte end-of-file");
	    if (!ch) berr ("4th byte unexpected byte code 0x00");
	    ulit |= (ch & 0x7f) << 21;
	    if (ch & 0x80) {
	      ch = read_next_binary_char ();
	      if (ch == EOF) berr ("5th byte end-of-file");
	      if (!ch) berr ("5th byte unexpected byte code 0x00");
	      if (ch > 0xf) berr ("5th byte code 0x%02x too large", ch);
	      ulit |= (ch & 0xf) << 28;
	    }
	  }
	}
      } else if (ulit == 1) berr ("1st byte unexpected byte code 0x01");
      lit = ulit >> 1;
      assert (lit >= 0);
      if (ulit) {
	if (COUNT (clause) == (long) UINT_MAX)
	  berr ("clause with more than UINT_MAX = %u literals", UINT_MAX);
	if (ulit & 1) lit = -lit;
	PUSH (clause, lit);
      } else {
	found_proof_clause (first == 'd');
	CLEAR (clause);
	parsed_clauses++;
	break;
      }
    }
  }
  *parsed_clauses_ptr = parsed_clauses;
}

static int force_non_binary;

static int try_parsing_binary_proof (int * parsed_clauses_ptr) {
  const int max_buffer_size = 32;
  int ch, res;
  if (force_non_binary) return 0;
  buffering = 1;
  ch = buffered_read_next_char ();
  if (ch == EOF) {
    msg1 ("assuming ASCII format after immediate end-of-file");
    res = 1;
  } if (ch == 'a') {
    msg1 ("first byte 'a' indicates binary proof format");
    res = 1;
  } else if (ch == 'd') {
    res = 0;
    while (!res && COUNT (buffer) < max_buffer_size) {
      ch = buffered_read_next_char ();
      if (ch == EOF) break;
      if (ch == 'd') continue;
      if (ch == ' ' || ch == '\n' || ch == '\r') continue;
      if (ch != '-' && !isdigit (ch)) res = 1;
    }
    if (res) {
      assert (0 <= ch), assert (ch < 256);
      if (isprint (ch))
	msg1 ("byte %d character '%c' indicates binary proof format",
	  (int) COUNT (buffer), ch);
      else
	msg1 ("byte %d code 0x%02x indicates binary proof format",
	  (int) COUNT (buffer), ch);
    } else {
      if (COUNT (buffer) < max_buffer_size)
	msg1 ("all %lu bytes fit ASCII format so assuming ASCII format",
	  (int) COUNT (buffer));
      else
	msg1 ("first %lu bytes fit ASCII format so assuming ASCII format",
	  (int) COUNT (buffer));
    }
  } else if (isprint (ch)) {
    msg1 ("first byte character '%c' indicates ASCII format", ch);
    res = 0;
  } else {
    msg1 ("first byte code 0x%02x forces binary format", ch);
    res = 1;
  }
  readchars = buffering = 0;
  if (!res) return 0;
  proof.binary = 1;
  parse_binary_proof (parsed_clauses_ptr);
  return 1;
}

#define CONTEXT proof
#define PARSE parse_proof
#define HEADER found_proof_header
#define CLAUSE found_proof_clause
#define BINARY try_parsing_binary_proof (&parsed_clauses)
#include "parse.h"

/*------------------------------------------------------------------------*/

static void found_post_cnf_header (int maxvar, int clauses) {
  assert (EMPTY (post.clauses));
  assert (!post.header.found);
  assert (maxvar >= 0);
  assert (clauses >= 0);
  post.header.found = 1;
  post.header.maxvar = maxvar;
  post.header.clauses = clauses;
}

static void found_post_cnf_clause (int deleted) {
  unsigned long idx;
  Cls * c;
  if (deleted) perr ("found 'd' clause in post-CNF");
  idx = COUNT (post.clauses);
  if (!(c = *find_clause_position_in_hash_table ())) {
    c = new_clause (&post);
    cdie (0, &post, idx, "was never added");
  } else {
    old_clause (&post, c);
    if (!c->refs)
      cdie (1, &post, idx, "has been deleted and is not alive");
  }
  cmsg2 (&post, idx, "parsed");
  check_empty (&post, idx);
}

#define CONTEXT post
#define PARSE parse_post_cnf
#define HEADER found_post_cnf_header
#define CLAUSE found_post_cnf_clause
#define BINARY 0
#include "parse.h"

/*------------------------------------------------------------------------*/

static void parse () {
  parse_pre_cnf ();
  parse_proof ();
  if (post.name) parse_post_cnf ();
}

/*------------------------------------------------------------------------*/

static Cls * empty, * conflict;
static signed char * vals;
static ClsPtrStk * occs;
static int level, next, root;
static STACK(int) trail;

static void init_propagation () {
  NEW (occs, 2*maxvar + 1);
  occs += maxvar;
  NEW (vals, maxvar + 1);
}

static void reset_propagation () {
  int lit;
  for (lit = -maxvar; lit <= maxvar; lit++)
    RELEASE (occs[lit]);
  occs -= maxvar;
  DEL (occs, 2*maxvar + 1);
  DEL (vals, maxvar + 1);
  RELEASE (trail);
}

static int val (int lit) {
  int idx = abs (lit), res;
  assert (idx), assert (idx <= maxvar);
  res = vals[idx];
  if (lit < 0) res = -res;
  return res;
}

static void assign (int lit, Cls * reason) {
  int idx = abs (lit);
  signed char v = (lit < 0) ? -1 : 1;
  assert (idx), assert (idx <= maxvar);
  assert (!vals[idx]);
  vals[idx] = v;
  if (verbosity > 1) {
    if (!level) msg2 ("unit %d", lit);
    else if (!reason) msg3 ("assume %d", lit);
    else msg3 ("assign %d", lit);
  }
  assert (val (lit) > 0);
  assert (val (-lit) < 0);
  PUSH (trail, lit);
}

static void unassign (int lit) {
  int idx = abs (lit);
  assert (idx), assert (idx <= maxvar);
  assert (vals[idx]);
  vals[idx] = 0;
  msg3 ("unassign %d", lit);
  assert (!val (lit));
}

static void check_watches () {
#ifndef NCHECK
  unsigned long i, count = 0;
  int lit;
  Cls * c;
  for (i = 0; i < size_clause_hash_table; i++)
    for (c = clause_hash_table[i]; c; c = c->chain)
      count++, c->watched = 0;
  assert (count == num_entries_clause_hash_table);
  for (lit = -maxvar; lit <= maxvar; lit++) {
    ClsPtrStk * s = occs + lit;
    int v0, v1;
    Cls ** p;
    for (p = s->start; p < s->top; p++) {
      c = *p;
      if (c->satisfied) continue;
      if (!c->refs) continue;
      assert (c->lits[0] == lit || c->lits[1] == lit);
      assert (c->watched < 2);
      c->watched++;
      if (conflict) continue;
      if (empty) continue;
      v0 = val (c->lits[0]);
      v1 = val (c->lits[1]);
      if (v0 < 0) assert (v1 > 0);
      if (v1 < 0) assert (v0 > 0);
    }
  }
#endif
}

#if MAXVERBOSITY > 1

static char clause_id_buffer[80];

static const char * clause_id (Cls * c) {
  if (!details) return "clause id[unknown]";
  sprintf (clause_id_buffer,
    "clause id[%lu]",
     clause_details (c)->id);
  return clause_id_buffer;
}

#endif

static void watch_literal (int lit, Cls * c) {
  assert (lit == c->lits[0] || lit == c->lits[1]);
  PUSH (occs[lit], c);
  msg3 ("watching literal %d in %s", lit, clause_id (c));
}

static void watch_clause (Cls * c) {
  int * lits = c->lits;
  unsigned pos, i = 2;
  for (pos = 0; pos < 2; pos++) {
    int lit = lits[pos], other;
    if (!val (lit)) continue;
    assert (i < c->size);
    while (val (other = lits[i]))
      i++, assert (i < c->size);
    lits[pos] = other, lits[i++] = lit;
  }
  assert (!val (lits[0]));
  assert (!val (lits[1]));
  watch_literal (lits[0], c);
  watch_literal (lits[1], c);
  check_watches ();
}

static void unwatch_literal (int lit, Cls * c) {
  ClsPtrStk * s = occs + lit;
  Cls ** p;
  for (p = s->start; p < s->top; p++)
    if (*p == c) break;
  if (p == s->top) return;
  while (p < s->top) p[0] = p[1], p++;
  assert (s->start < s->top);
  s->top--;
  msg3 ("unwatched literal %d in %s", lit, clause_id (c));
}

static void unwatch_clause (Cls * c) {
  assert (c->size > 1);
  unwatch_literal (c->lits[0], c);
  unwatch_literal (c->lits[1], c);
}

static void backtrack () {
  assert (level);
  while (COUNT (trail) > root) {
    int lit = *--trail.top;
    unassign (lit);
  }
  conflict = 0;
  next = root;
  level = 0;
}

static void mark_clause_as_permanently_satisfied (Cls * c) {
  assert (!level);
  assert (!c->satisfied);
  c->satisfied = 1;
  msg2 ("marked %s as permanently satisfied", clause_id (c));
}

static int propagate () {
  assert (!empty);
  while (!conflict && next < COUNT (trail)) {
    int lit = -trail.start[next++];
    Cls ** p, ** q;
    ClsPtrStk * s;
    INC (propagations);
    msg3 ("propagating %d", -lit);
    s = occs + lit;
    p = q = s->start;
    while (p < s->top) {
      unsigned pos, i, size;
      int * lits, other, v;
      Cls * c = *p++;
      if (c->satisfied) continue;
      if (!c->refs) continue;
      size = c->size;
      lits = c->lits;
      *q++ = c;
      assert (size > 1);
      pos = (lits[1] == lit);
      assert (lits[pos] == lit);
      for (i = 2; i < size; i++) {
	v = val (other = lits[i]);
	if (v < 0) continue;
	if (v > 0 && !level) goto CLAUSE_SATISFIED;
	lits[pos] = other;
	lits[i] = lit;
	watch_literal (other, c);
	q--;
	goto CONTINUE_OUTER_LOOP;
      }
      v = val (other = lits[!pos]);
      if (v < 0) {
	msg3 ("found conflict");
	conflict = c;
	break;
      } else if (!v) { 
	assign (other, c);
	if (!level) goto CLAUSE_SATISFIED;
      } else {
	assert (v > 0);
	if (!level) {
CLAUSE_SATISFIED:
	  mark_clause_as_permanently_satisfied (c);
	  q--;
	}
      }
CONTINUE_OUTER_LOOP:
      ;
    }
    while (p < s->top) *q++ = *p++;
    s->top = q;
  }
  if (!level && conflict) {
    assert (conflict);
    msg1 ("root-level propagation produced empty clause");
    empty = conflict;
    conflict->empty = 1;
  }
  check_watches ();
  return !conflict;
}

/*------------------------------------------------------------------------*/

static void
disconnect_clause (Cls * c, const char * type, unsigned long idx) {
  assert (!plain_mode);
  if (!c->refs) { assert (c->satisfied); return; }
  assert (c->size > 1);
  assert (c->refs > 0);
  if (!--c->refs) unwatch_clause (c);
  cmsg2 (&proof, idx, "disconnected with %lu references", c->refs);
  if (details && !c->refs)
    clause_details (c)->del = position (current_context, idx);
}

static void
connect_clause (Cls * c, const char * type, unsigned long idx) {
  int unit = 0;
  unsigned i;
  assert (!level);
  if (c->satisfied) return;
  if (c->empty) { assert (empty); return; }
  for (i = 0; i < c->size; i++) {
    int lit = c->lits[i];
    int v = val (lit);
    if (v > 0) break;
    if (v < 0) continue;
    if (unit) unit = INT_MIN;
    else unit = lit;
  }
  if (i < c->size) {
    msg2 ("ignoring root-level satisfied clause");
    c->satisfied = 1;
  } else if (!unit) {
    c->empty = 1;
    if (empty) msg2 ("this is another empty root-level clause");
    else {
      msg1 ("found an empty root-level clause");
      empty = c;
    }
  } else if (unit == INT_MIN) {
    if (c->refs) msg2 ("will not watch this clause again");
    else watch_clause (c);
    c->refs++, assert (c->refs);
    if (details && c->refs == 1)
      clause_details (c)->add = position (current_context, idx);
  } else {
    msg2 ("found new root-level unit %d", unit);
    assign (unit, 0);
    if (propagate ()) c->satisfied = 1;
  }
#if MAXVERBOSITY > 1
  assert (current_context);
  cmsg2 (current_context, idx, "connected with %lu references", c->refs);
#endif
}

static void
check_clause (Cls * c, const char * type, unsigned long idx) {
  if (empty) return;
  if (c->satisfied) return;
  if (c->refs > 0) return;
  INC (checks);
  assert (!c->refs);
  assert (!conflict);
  assert (!level);
  if (!propagate ()) {
    assert (!empty);
    assert (conflict);
    msg1 ("root-level propagation produced empty clause");
    empty = conflict;
    conflict->empty = 1;
  } else {
    const int * lits = c->lits;
    unsigned i, size = c->size;
    for (i = 0; i < size; i++) {
      int lit = lits[i], v = val (lit);
      if (v > 0) break;
    }
    if (i < size) mark_clause_as_permanently_satisfied (c);
    else {
      root = COUNT (trail);
      assert (next == root);
      level = 1;
      for (i = 0; i < size; i++) {
	int lit = lits[i], v = val (lit);
	assert (v <= 0);
	if (v < 0) continue;
	assign (-lit, 0);
	INC (decisions);
      }
      if (propagate ()) cdie (0, &proof, idx, "is not implied");
      backtrack ();
    }
  }
  cmsg2 (&proof, idx, "checked");
}

/*------------------------------------------------------------------------*/

static int sat_mode, plain_mode;
static int forward_checking = 1;

static void forward () {
  unsigned long i;
  Cls ** p, * c;
  char * q;
#if MAXVERBOSITY > 0
  Timer timer;
  start_timer (1, &timer);
#endif
  msg1 ("starting complete forward checking");
  for (i = 0; i < size_clause_hash_table; i++)
    for (c = clause_hash_table[i]; c; c = c->chain)
      c->refs = 0;
#if MAXVERBOSITY > 1
  current_context = &pre;
#endif
  for (p = pre.clauses.start, i = 0;
       !empty && p < pre.clauses.top;
       p++, i++)
    connect_clause (*p, "pre", i);
#if MAXVERBOSITY > 1
  current_context = &proof;
#endif
  for (p = proof.clauses.start, q = proof.deleted.start, i = 0;
       !empty && p < proof.clauses.top;
       p++, q++, i++) {
    if (!*q) check_clause (*p, "proof", i), connect_clause (*p, "proof", i);
    else if (!plain_mode) disconnect_clause (*p, "proof", i);
  }
#if MAXVERBOSITY > 0
  stop_timer (&timer, "finished forward checking in");
#endif
}

static void backward () {
  die ("partial backward checking not implemented yet");
}

static void check () {
  msg1 ("");
  init_propagation ();
  if (forward_checking) {
    forward ();
    if (!post.name &&
        !pre.empty.found && !proof.empty.found &&
	!sat_mode)
      die ("no post-CNF, no empty clause(s), no SAT-mode");
  } else backward ();
}

/*------------------------------------------------------------------------*/

static void stats () {
#if MAXVERBOSITY > 0
  if (verbosity <= 0) return;
  printf ("c\n");
  printf ("c found %lu unique clauses in total\n", clause_ids);
  printf ("c %lu hash table searches with %lu collisions (average %.2f)\n",
    searches_clause_hash_table, collisions_clause_hash_table,
    average (collisions_clause_hash_table, searches_clause_hash_table));
  printf ("c maximum allocated memory %.1f MB\n",
    maxbytes/(double)(1ul<<20));
  printf ("c %lu checks, %lu decisions, %lu propagations\n",
    checks, decisions, propagations);
  fflush (stdout);
#endif
}

/*------------------------------------------------------------------------*/

#define CHECK_SIZEOF(TYPE,BYTES) \
do { \
  if (sizeof (TYPE) == BYTES) break; \
  die ("sizeof (%s) is %d but expected %d", \
    #TYPE, sizeof (TYPE), BYTES); \
} while (0)

#if MAXVERBOSITY > 0
static Timer global_timer;
#endif

static void init () {
#if MAXVERBOSITY > 0
  start_real_time = current_real_time ();
#endif
  if (sizeof (void *) != 8)
    die ("only supported for 64-bit compilation'");
  CHECK_SIZEOF (int, 4);
  CHECK_SIZEOF (unsigned, 4);
  CHECK_SIZEOF (long, 8);
  CHECK_SIZEOF (unsigned long, 8);
  pre.type = -1;
  assert (!proof.type);
  post.type = 1;
}

static void delete_all_clauses () {
  unsigned long i;
  Cls * c, * n;
  for (i = 0; i < size_clause_hash_table; i++) {
    for (c = clause_hash_table[i]; c; c = n) {
      n = c->chain;
      delete_clause (c);
    }
  }
}

static void reset_context (Context * context) {
  RELEASE (context->clauses);
  RELEASE (context->position);
  RELEASE (context->deleted);
}

static void reset () {
  reset_propagation ();
  delete_all_clauses ();
  DEL (clause_hash_table, size_clause_hash_table);
  release_buffer (&file_info_buffer);
  RELEASE (line);
  RELEASE (buffer);
  RELEASE (clause);
  reset_context (&pre);
  reset_context (&proof);
  reset_context (&post);
  assert (getenv ("LEAK") || !allocated);
#if MAXVERBOSITY > 0
  stop_timer (&global_timer, "total time");
#endif
  msg1 ("");
}

/*------------------------------------------------------------------------*/

#include "config.h"

static void banner () {
  msg1 ("DRABT Proof Checker Version %s", VERSION);
  msg1 ("Copyright (C) 2015, Armin Biere, JKU");
  msg1 ("Compiled with '%s'", COMPILE);
  msg1 ("");
}

static void adjust_file_info_buffer () {
  size_t len = strlen (pre.name), tmp;
  tmp = strlen (proof.name);
  if (tmp > len) len = tmp;
  if (post.name) {
    tmp = strlen (post.name);
    if (tmp > len) len = tmp;
  }
  adjust_buffer (&file_info_buffer, len + 20);
}

static void setup () {
  assert (verbosity <= MAXVERBOSITY);
  msg1 ("verbosity %d (compile time maximum %d)", verbosity, MAXVERBOSITY);

  if (ignore_warnings) msg1 ("warnings are silently ignored");
  if (pedantic) msg1 ("pedantic mode (warnings treated as errors)");
  if (sloppy) msg1 ("sloppy mode (ignore some 'ignorable' errors)");

  msg1 ("SAT mode %s", sat_mode ? "enabled" : "disabled");
  if (sat_mode) forward_checking = 1;
  msg1 ("forward checking %s",
    forward_checking ? "enabled" : "disabled");

  if (!post.name && !sat_mode)
    msg1 ("checking proof to be refutation (no post-CNF, no SAT-mode)");
  else if (post.name) {
    msg1 ("checking all post-CNF clauses are derived");
    if (post.empty.found)
      msg1 ("checking empty clause in post-CNF (proof is refutation)");
    else
      msg1 ("not checking proof to be refutation "
            "(no empty clause in post-CNF)");
  } else {
    assert (sat_mode);
    assert (forward_checking);
    msg1 ("checking all individual proof steps");
    msg1 ("not checking proof to be a refutation (no post-CNF)");
  }

  if (more_details) details = 1;
  else if (less_details) details = 0;
  else details = (verbosity > 0);

  if (details) msg1 ("error messages with full details (needs more memory)");
  else msg1 ("error messages with less details (needs less memory)");

  if (!proof.name) proof.read_from_stdin = 1, proof.name = stdin_name;

  adjust_file_info_buffer ();
}

static int exists_file (const char * name) {
  struct stat buf;
  return !stat (name, &buf);
}

static void usage () {
  printf (USAGE, MAXVERBOSITY);
  exit (1);
}

static void args (int argc, char ** argv) {
  int i;
  for (i = 1; i < argc; i++) {
         if (!strcmp (argv[i], "-h")) usage ();
    else if (!strcmp (argv[i], "-v")) verbosity++;
    else if (!strcmp (argv[i], "-f")) forward_checking = 1;
    else if (!strcmp (argv[i], "-b")) forward_checking = 0;
    else if (!strcmp (argv[i], "-p")) plain_mode = 1;
    else if (!strcmp (argv[i], "-S")) sat_mode = 1;
    else if (!strcmp (argv[i], "--force-non-binary")) force_non_binary = 1;
    else if (!strcmp (argv[i], "-I") ||
             !strcmp (argv[i], "--ignore-warnings"))
      ignore_warnings = 1;
    else if (!strcmp (argv[i], "-M") ||
             !strcmp (argv[i], "--more-details")) more_details = 1;
    else if (!strcmp (argv[i], "-L") ||
             !strcmp (argv[i], "--less-details")) less_details = 1;
    else if (!strcmp (argv[i], "-P") ||
             !strcmp (argv[i], "--pedantic")) pedantic = 1;
    else if (!strcmp (argv[i], "-s") ||
             !strcmp (argv[i], "--sloppy")) sloppy = 1;
    else if (argv[i][0] == '-')
      die ("invalid option '%s' (try '-h')", argv[i]);
    else if (post.name)
      die ("too many files specified (try '-h')");
    else {
      Context * context;
           if (!pre.name)   context = &pre;
      else if (!proof.name) context = &proof;
      else                  context = &post;
      if (!exists_file (argv[i]))
	die ("%s file '%s' does not exist (try '-h')",
	  context_type (context), argv[i]);
      context->name = argv[i];
    }
  }
  if (!pre.name) usage ();
  if (sloppy && pedantic)
    die ("can not use both '--pedantic' and '--sloppy'");
  if (less_details && more_details)
    die ("can not use both '--less-details' and '--more-details'");
  if (verbosity > MAXVERBOSITY)
    die ("verbose level %d times '-v' too large (compile time maximum %d)",
      verbosity, MAXVERBOSITY);
#if MAXVERBOSITY > 0
  start_timer (1, &global_timer);
#endif
}

int main (int argc, char ** argv) {
  init ();
  args (argc, argv);
  banner ();
  setup ();
  parse ();
  check ();
  stats ();
  reset ();
  printf ("s VERIFIED\n");
  return 0;
}
